from __future__ import annotations

import datetime
from typing import Any, Callable, Type, get_type_hints

import numpy as np
import pandas as pd
from pydantic import BaseModel, StrictBool, StrictFloat, StrictInt, StrictStr


class Record(BaseModel):
    class Config:
        arbitrary_types_allowed = True
        extra = "forbid"
        validate_assignment = True

    def series(self) -> pd.Series[Any]:
        return pd.Series(self.dict())

    @classmethod
    @property
    def dtypes(cls) -> dict[str, Type[Any]]:
        return get_type_hints(cls)


class Index(Record):
    experiment_name: StrictStr
    run_id: StrictInt


class SolverInfo(Record):
    experiment_name: StrictStr
    run_id: StrictInt
    solver_name: StrictStr
    source: StrictStr
    solver_id: StrictInt


class SolverContent(Record):
    solver_id: StrictInt
    content: Callable[..., Any]


class ParameterInfo(Record):
    experiment_name: StrictStr
    run_id: StrictInt
    parameter_name: StrictStr
    parameter_id: StrictInt


class ParameterContent(Record):
    parameter_id: StrictInt
    content: Any


class ResultInfo(Record):
    experiment_name: StrictStr
    run_id: StrictInt
    result_name: StrictStr
    result_id: StrictInt


class ResultContent(Record):
    """ResultValue

    Attributes:
        value_id (int): value id
        value (Any): value
    """

    result_id: StrictInt
    content: Any


def get_pandas_dtypes(dtypes: dict[str, Type[Any]]) -> dict[str, str]:
    """isinstance

    Args:
        value (Any): value
        expected_type (Type[T]): expected type

    Returns:
        T: value
    """

    uint_types = [np.uint8, np.uint16, np.uint32, np.uint64]
    int_types = [int, np.int8, np.int16, np.int32, np.int64]
    float_types = [float, np.float16, np.float32, np.float64]
    complex_types = [complex, np.complex64, np.complex128]

    pandas_dtypes: dict[str, str] = {}
    for k, v in dtypes.items():
        if v is StrictInt:
            pandas_dtypes[k] = "int"
        elif v is StrictFloat:
            pandas_dtypes[k] = "float"
        elif v in (bool, StrictBool):
            pandas_dtypes[k] = "boolean"
        elif v in (str, StrictStr):
            pandas_dtypes[k] = "string"
        elif v in (datetime.date, datetime.datetime):
            pandas_dtypes[k] = "datetime64[ns]"
        elif v is datetime.timedelta:
            pandas_dtypes[k] = "timedelta64[ns]"
        elif v in uint_types + int_types + float_types + complex_types:
            pandas_dtypes[k] = v.__name__
        else:
            pandas_dtypes[k] = "object"
    return pandas_dtypes


def get_simple_dypes(dtypes: dict[str, str]) -> dict[str, Type[Any]]:
    uint_types = ["uint8", "uint16", "uint32", "uint64"]
    int_types = ["int8", "int16", "int32", "int64"]
    float_types = ["float16", "float32", "float64"]
    complex_types = ["complex64", "complex128"]
    bool_types = ["bool", "boolean"]

    simple_dtypes: dict[str, Type[Any]] = {}
    for k, v in dtypes.items():
        if v is uint_types:
            simple_dtypes[k] = np.uint64
        elif v in int_types:
            simple_dtypes[k] = int
        elif v in float_types:
            simple_dtypes[k] = float
        elif v in complex_types:
            simple_dtypes[k] = complex
        elif v in bool_types:
            simple_dtypes[k] = bool
        elif v == "string":
            simple_dtypes[k] = str
        elif v == "datetime64[ns]":
            simple_dtypes[k] = datetime.datetime
        elif v == "timedelta64[ns]":
            simple_dtypes[k] = datetime.timedelta
        else:
            simple_dtypes[k] = Any
    return simple_dtypes



from __future__ import annotations

import inspect
import json
import pathlib
import types
import uuid
from typing import Any, Callable, Literal, Optional, TypedDict

import h5py
import jijmodeling as jm
import numpy as np
import pandas as pd
from jijzept.response import JijModelingResponse
from pandas import DataFrame

from minto.consts.default import DEFAULT_RESULT_DIR
from minto.records.records import (
    Index,
    ParameterContent,
    ParameterInfo,
    ResultContent,
    ResultInfo,
    SolverContent,
    SolverInfo,
)
from minto.records.sampleset_expansion import expand_sampleset
from minto.table.table import SchemaBasedTable
from minto.utils.rc_sampleset import SampleSet, from_old_sampleset, serialize_sampleset


class DatabaseComponentSchema(TypedDict):
    info: SchemaBasedTable
    content: SchemaBasedTable


class DatabaseSchema(TypedDict):
    index: SchemaBasedTable
    solver: DatabaseComponentSchema
    parameter: DatabaseComponentSchema
    result: DatabaseComponentSchema


class Experiment:
    def __init__(
        self,
        name: Optional[str] = None,
        savedir: str | pathlib.Path = DEFAULT_RESULT_DIR,
    ):
        self.name = name or str(uuid.uuid4())
        self.savedir = pathlib.Path(savedir)

        database: DatabaseSchema = {
            "index": SchemaBasedTable(Index.dtypes),
            "solver": {
                "info": SchemaBasedTable(SolverInfo.dtypes),
                "content": SchemaBasedTable(SolverContent.dtypes),
            },
            "parameter": {
                "info": SchemaBasedTable(ParameterInfo.dtypes),
                "content": SchemaBasedTable(ParameterContent.dtypes),
            },
            "result": {
                "info": SchemaBasedTable(ResultInfo.dtypes),
                "content": SchemaBasedTable(ResultContent.dtypes),
            },
        }
        object.__setattr__(self, "database", database)

    def __enter__(self) -> Experiment:
        """Set up Experiment.
        Automatically makes a directory for saving the experiment, if it doesn't exist.
        """
        _mkdir(self)

        database: DatabaseSchema = getattr(self, "database")

        run_id = int(database["index"][-1].series()["run_id"])
        solver_id = len(database["solver"]["info"])

        return self

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        """Saves the experiment if autosave is True."""
        pass

    def run(self) -> Experiment:
        """Run the experiment."""
        database: DatabaseSchema = getattr(self, "database")

        if database["index"].empty():
            run_id = 0
        else:
            run_id = database["index"][-1].series()["run_id"] + 1
        database["index"].insert(Index(experiment_name=self.name, run_id=run_id))

        return self

    def table(
        self,
        key: Literal["solver", "parameter", "result"] | None = None,
        enable_sampleset_expansion: bool = True,
    ) -> pd.DataFrame:
        """Merge the experiment table and return it as a DataFrame.

        Returns:
            pd.DataFrame: The merged table.
        """
        database: DatabaseSchema = getattr(self, "database")

        solver_df = _get_component_dataframe(self, "solver")
        if key == "solver":
            return solver_df
        parameter_df = _get_component_dataframe(self, "parameter")
        if key == "parameter":
            return parameter_df
        result_df = _get_component_dataframe(self, "result")
        if key == "result":
            return result_df

        df = database["index"].dataframe()
        # Merge solver
        if not solver_df.empty:
            df = df.merge(
                _pivot(solver_df, columns="solver_name", values="content"),
                on=["experiment_name", "run_id"],
                how="outer",
            )

        # Merge parameter
        if not parameter_df.empty:
            df = df.merge(
                _pivot(parameter_df, columns="parameter_name", values="content"),
                on=["experiment_name", "run_id"],
                how="outer",
            )

        # Merge result
        if not result_df.empty:
            df = df.merge(
                _pivot(result_df, columns="result_name", values="content"),
                on=["experiment_name", "run_id"],
                how="outer",
            )

        # Expand sampleset
        if enable_sampleset_expansion:
            sampleset_df = expand_sampleset(database["result"]["content"].dataframe())
            if not sampleset_df.empty:
                sampleset_df = pd.merge(
                    database["result"]["info"].dataframe()[
                        ["experiment_name", "run_id", "result_id"]
                    ],
                    sampleset_df,
                    on="result_id",
                    how="outer",
                ).drop(columns="result_id")

                result_names = [
                    name
                    for name in result_df["result_name"].unique()
                    if isinstance(
                        result_df[result_df["result_name"] == name]["content"].iloc[0],
                        (SampleSet, jm.SampleSet, JijModelingResponse),
                    )
                ]

                df = df.merge(sampleset_df, on=["experiment_name", "run_id"]).drop(
                    columns=result_names
                )
        return df

    def log_solver(self, name: str, solver: Callable[..., Any]) -> None:
        database: DatabaseSchema = getattr(self, "database")

        run_id = int(database["index"][-1].series()["run_id"])
        solver_id = len(database["solver"]["info"])

        if isinstance(solver, types.FunctionType):
            source = inspect.getfile(solver)
        else:
            if _is_running_in_notebook():
                source = "Dynamically generated in Jupyter Notebook"
            else:
                if isinstance(solver, types.MethodType):
                    source = inspect.getfile(solver)
                else:
                    source = inspect.getfile(solver.__class__)

        info = SolverInfo(
            experiment_name=self.name,
            run_id=run_id,
            solver_name=name,
            source=source,
            solver_id=solver_id,
        )
        content = SolverContent(solver_id=solver_id, content=solver)

        database["solver"]["info"].insert(info)
        database["solver"]["content"].insert(content)

    def log_parameter(self, name: str, parameter: Any) -> None:
        """Log a parameter to the experiment.

        Args:
            parameter (Parameter): The parameter to be logged.
        """
        database: DatabaseSchema = getattr(self, "database")

        run_id = int(database["index"][-1].series()["run_id"])
        parameter_id = len(database["parameter"]["info"])

        info = ParameterInfo(
            experiment_name=self.name,
            run_id=run_id,
            parameter_name=name,
            parameter_id=parameter_id,
        )
        content = ParameterContent(parameter_id=parameter_id, content=parameter)

        database["parameter"]["info"].insert(info)
        database["parameter"]["content"].insert(content)

    def log_result(self, name: str, result: Any) -> None:
        database: DatabaseSchema = getattr(self, "database")

        run_id = int(database["index"][-1].series()["run_id"])
        result_id = len(database["result"]["info"])

        info = ResultInfo(
            experiment_name=self.name,
            run_id=run_id,
            result_name=name,
            result_id=result_id,
        )
        content = ResultContent(result_id=result_id, content=result)

        database["result"]["info"].insert(info)
        database["result"]["content"].insert(content)

    def save(self) -> None:
        """Save the experiment to a file."""
        database: DatabaseSchema = getattr(self, "database")

        base_dir = self.savedir / self.name

        keys: list[Literal["index", "solver", "parameter", "result"]] = [
            "index",
            "solver",
            "parameter",
            "result",
        ]
        dtypes = {}
        for key in keys:
            dtypes[key] = {"info": {}, "content": {}}

        for key in keys:
            if key == "index":
                database[key].dataframe().to_csv(base_dir / f"{key}.csv", index=False)
                dtypes[key] = database[key].pandas_dtypes
            else:
                data_dir = base_dir / key

                # get dtypes
                dtypes[key]["info"] = database[key]["info"].pandas_dtypes
                dtypes[key]["content"] = database[key]["content"].pandas_dtypes

                # save info
                database[key]["info"].dataframe().to_csv(
                    data_dir / "info.csv", index=False
                )

                # save content
                _save_content(database, key, base_dir)

        with open(base_dir / "dtypes.json", "w") as f:
            json.dump(dtypes, f)


class _NumpyEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, np.ndarray):
            return o.tolist()
        return super().default(o)


def _mkdir(experiment: Experiment) -> None:
    for key in ["solver", "parameter", "result"]:
        d = experiment.savedir / experiment.name / key
        d.mkdir(parents=True, exist_ok=True)

        if key in ["parameter", "result"]:
            problem_dir = d / "problems"
            problem_dir.mkdir(parents=True, exist_ok=True)

            sampleset_dir = d / "samplesets"
            sampleset_dir.mkdir(parents=True, exist_ok=True)


def _get_component_dataframe(
    experiment: Experiment, key: Literal["solver", "parameter", "result"]
) -> DataFrame:
    database: DatabaseSchema = getattr(experiment, "database")

    return pd.merge(
        database[key]["info"].dataframe(),
        database[key]["content"].dataframe(),
        on=f"{key}_id",
    )


def _pivot(df: DataFrame, columns: str | list[str], values: str) -> DataFrame:
    return df.pivot_table(
        index=["experiment_name", "run_id"],
        columns=columns,
        values=values,
        aggfunc=lambda x: x,
        dropna=False,
    ).reset_index()


def _is_running_in_notebook():
    try:
        ipython = get_ipython()
        # Jupyter Notebook or JupyterLab
        if "IPKernelApp" in ipython.config:
            return True
    except NameError:
        return False


def _save_content(
    database: DatabaseSchema,
    key: Literal["solver", "parameter", "result"],
    base_dir: pathlib.Path,
) -> None:
    for index, record in database[key]["content"].dict().items():
        content_id = record[f"{key}_id"]
        content = record["content"]

        with h5py.File(base_dir / f"{key}" / "content.h5", "a") as f:
            # save problem
            if isinstance(content, jm.Problem):
                with open(
                    base_dir / f"{key}" / "problems" / f"{content_id}.pb", "wb"
                ) as f:
                    f.write(jm.to_protobuf(content))

            # save jijzept JijModelingResponse
            elif isinstance(content, JijModelingResponse):
                sampleset = serialize_sampleset(from_old_sampleset(content.sample_set))
                with open(
                    base_dir / f"{key}" / "samplesets" / f"{content_id}.json", "w"
                ) as f:
                    json.dump(sampleset, f)

            # save jijmodeling SampleSet
            elif isinstance(content, (SampleSet, jm.SampleSet)):
                sampleset = serialize_sampleset(from_old_sampleset(content))
                with open(
                    base_dir / f"{key}" / "samplesets" / f"{content_id}.json", "w"
                ) as f:
                    json.dump(sampleset, f)

            # save other
            else:
                if str(index) in f:
                    group = f[str(index)]
                else:
                    group = f.create_group(str(index))

                for name, value in record.items():
                    if name in group:
                        del group[name]

                    if key == "solver" and name == "content":
                        value = database[key]["info"].dict()[index]["source"]

                    value = json.dumps(value, cls=_NumpyEncoder)
                    group.create_dataset(name, data=value)



from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)


import minto.utils as utils
from minto.experiment.experiment import Experiment
from minto.io.load import load
from minto.table.table import SchemaBasedTable

__all__ = ["load", "Experiment", "SchemaBasedTable", "utils"]



from __future__ import annotations

import dataclasses
import enum
import typing as typ
import uuid
from typing import Any

import jijmodeling as jm
import numpy as np


@dataclasses.dataclass
class SolvingTime:
    compiling_time: float = 0.0
    transpiling_time: float = 0.0
    preprocess: float = 0.0
    solving_time: float = 0.0
    decoding_time: float = 0.0
    postprocess: float = 0.0

    def total(self) -> float:
        total_time = sum(dataclasses.asdict(self).values())
        return total_time


@dataclasses.dataclass
class SystemTime:
    posting_time: typ.Optional[float] = None
    request_queuing_time: typ.Optional[float] = None
    fetching_problem_time: typ.Optional[float] = None
    fetching_result_time: typ.Optional[float] = None
    deserialize_time: typ.Optional[float] = None


@dataclasses.dataclass
class MeasuringTime:
    solving_time: SolvingTime = SolvingTime()
    system_time: SystemTime = SystemTime()

    def total(self) -> float:
        total_time = self.solving_time.total()
        for value in dataclasses.asdict(self.system_time).values():
            if value is not None:
                total_time += value
        return total_time


Serializable = typ.Union[str, int, float, dict, list]


class VarType(enum.IntEnum):
    CONTINUOUS = 0
    INTEGER = 1
    BINARY = 2


@dataclasses.dataclass(frozen=True)
class VariableSparseValue:
    name: str
    value: dict[tuple[int, ...], float]
    shape: tuple[int, ...]
    vartype: VarType

    def to_dense(self) -> np.ndarray:
        dense_value = np.zeros(self.shape, dtype=np.float64)
        for key, value in self.value.items():
            dense_value[key] = value
        return dense_value

    @classmethod
    def from_array(
        cls, name: str, array: np.ndarray, vartype: VarType = VarType.CONTINUOUS
    ) -> "VariableSparseValue":
        # Check array is scalar or not
        if array.ndim == 0:
            nnz_value = np.array([array])
            subscripts = [[]]
        else:
            nnz_indices = np.nonzero(array)
            subscripts = np.array(list(nnz_indices)).T
            nnz_value = array[nnz_indices]
        return cls(
            name=name,
            value={tuple(key): value for key, value in zip(subscripts, nnz_value)},
            shape=array.shape,
            vartype=vartype,
        )


@dataclasses.dataclass
class ExprEvaluation:
    name: str
    total_violation: float
    expr_values: dict[tuple[int, ...], float]


@dataclasses.dataclass
class EvaluationResult:
    energy: float
    objective: float
    constraints: dict[str, ExprEvaluation]
    penalties: dict[str, ExprEvaluation]


@dataclasses.dataclass
class Sample:
    """The Sample class is designed to store the result of mathematical optimization.

    In the following, we explain the attributes of the Sample class in detail.
    `.vars` is stored only non-zero elements of the decision variables.
    For example, if the values are x = `[[0, 1, 2], [1, 0, 0]]`, it is stored as `{(0, 1): 1, (0, 2): 2, (1, 0): 1}`.
    If you want to get the dense array of the decision variables, use `.to_dense()` method.

    `.run_id` is a unique identifier for the run in which the solution was found. Note that this is not a unique identifier for the sample.
    The same sample can have same run_id if the same sample is obtained in the same run.
    For example, if you run heuristic solver 10 times, you can get 10 samples with the same run_id; You have to set the same run_id manually.
    However, if you run heuristic solver 10 times with different parameters, you can get 10 samples with different run_id.

    Attribuites:
        vars (dict[str, VariableSparseValue]): Stores the optimal values of decision variables in a sparse matrix format. Only non-zero elements are stored.
        evaluation_result (EvaluationResult): Stores the evaluation result of the objective function's value and constraints.
        num_occurrences (int): The number of occurrences of the sample.
        run_id (str): Identifier for the run in which the solution was found. Note that this is not a unique identifier for the sample.
        meta_info (dict[str, Serializable]): Stores the meta information of the sample.
    """

    vars: dict[str, VariableSparseValue]
    evaluation_result: EvaluationResult
    num_occurrences: int = 1
    run_id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))
    meta_info: dict[str, Serializable] = dataclasses.field(default_factory=dict)

    def is_feasible(self, epsilon=1e-10) -> bool:
        total_violation = sum(
            [v.total_violation for v in self.evaluation_result.constraints.values()]
        )
        return total_violation < epsilon

    def to_dense(self) -> dict[str, np.ndarray]:
        dense_vars = {}
        for name, var in self.vars.items():
            dense_vars[name] = var.to_dense()
        return dense_vars

    @classmethod
    def from_array(
        cls,
        array: dict[str, np.ndarray],
        num_occurances: int = 1,
        vartype: typ.Optional[dict[str, VarType]] = None,
        run_id: typ.Optional[str] = None,
        meta_info: typ.Optional[dict] = None,
    ) -> Sample:
        if vartype is None:
            vartype = {name: VarType.CONTINUOUS for name in array.keys()}
        if run_id is None:
            run_id = str(uuid.uuid4())
        if meta_info is None:
            meta_info = {}
        sparse_vars = {}
        for name, value in array.items():
            _vartype = vartype.get(name, VarType.CONTINUOUS)
            sparse_vars[name] = VariableSparseValue.from_array(name, value, _vartype)
        return cls(
            vars=sparse_vars,
            evaluation_result=EvaluationResult(0.0, 0.0, {}, {}),
            num_occurrences=num_occurances,
            run_id=run_id,
            meta_info=meta_info,
        )


@dataclasses.dataclass
class SampleSet:
    """SampleSet is a set of samples

    The SampleSet class is designed to store a collection of Sample instances.

    It behaves similar to a list of Samples with additional features and capabilities.
    - Stores metadata for each SampleSet. During concatenation of multiple SampleSets using the concat method, it maintains the metadata of each SampleSet.
    - Uses the run_id of each Sample as a key in a dictionary to store and retain the metadata of each SampleSet in run_info.
    - Provides an interface to perform searches on the set of Samples, such as retrieving the set of executable solutions, the solution with the minimum objective function value, or the solution with the minimum constraint violation.
    - Implements __len__, __getitem__, and __iter__ to behave like a regular list of Samples.

    Attributes:
        data (list[Sample]): List of samples
        set_id (str): ID of the sample set
        set_info (dict[str, Serializable]): Information of the sample set
        run_info (dict[str, Serializable]): Information of the run
        measuring_time (MeasuringTime): Measuring time of the sample set
        run_times (dict[str, MeasuringTime]): Measuring time of each run

    """

    data: list[Sample]
    set_id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))
    set_info: dict[str, Serializable] = dataclasses.field(default_factory=dict)
    run_info: dict[str, Serializable] = dataclasses.field(default_factory=dict)
    measuring_time: MeasuringTime = dataclasses.field(default_factory=MeasuringTime)
    run_times: dict[str, MeasuringTime] = dataclasses.field(default_factory=dict)

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, idx: int) -> Sample:
        return self.data[idx]

    def __iter__(self) -> typ.Iterator[Sample]:
        return iter(self.data)

    def feasibles(self, epsilon=1e-8) -> SampleSet:
        """Return a sampleset with only feasible samples

        Args:
            epsilon (float, optional): Tolerance for constraint violation. Defaults to 1e-8.

        Returns:
            SampleSet: A sampleset with only feasible samples
        """
        sampleset = SampleSet(
            [sample for sample in self.data if sample.is_feasible(epsilon)]
        )
        sampleset.set_info = self.set_info
        sampleset.run_info = self.run_info
        return sampleset

    def separate(self) -> dict[str, SampleSet]:
        """Separate the sampleset into different samplesets based on the run_id"""
        separated: dict[str, SampleSet] = {}
        for sample in self.data:
            if sample.run_id not in separated:
                separated[sample.run_id] = SampleSet([])
            separated[sample.run_id].data.append(sample)

        for run_id in separated.keys():
            separated[run_id].set_info = self.set_info
            separated[run_id].measuring_time = self.measuring_time
            if run_id in self.run_info:
                separated[run_id].run_info[run_id] = self.run_info[run_id]
            if run_id in self.run_times:
                separated[run_id].measuring_time = self.run_times[run_id]

        return separated

    @classmethod
    def concat(cls, family: list[SampleSet]) -> SampleSet:
        """Concatenate a list of samplesets into one

        Args:
            family (list[SampleSet]): A list of samplesets

        Returns:
            SampleSet: A sampleset that contains all samples from the input samplesets
        """
        data: list[Sample] = []
        run_info: dict[str, Serializable] = {}
        run_times: dict[str, MeasuringTime] = {}
        for sampleset in family:
            for _sampleset in sampleset.separate().values():
                data.extend(_sampleset.data)
                run_info.update(_sampleset.run_info)
                run_times.update(_sampleset.run_times)
        return cls(data, set_info={}, run_info=run_info)

    @classmethod
    def from_array(
        cls, samples: list[dict[str, typ.Union[np.ndarray, list]]]
    ) -> "SampleSet":
        data: list[Sample] = []
        run_id = str(uuid.uuid4())
        for sample in samples:
            _sample = {name: np.array(value) for name, value in sample.items()}
            data.append(Sample.from_array(_sample, run_id=run_id))
        return cls(data=data)


def serialize_sampleset(samplset: SampleSet) -> dict[str, Any]:
    obj = dataclasses.asdict(samplset)

    for sampleset_field_name, sampleset_field in obj.items():
        if sampleset_field_name == "data":
            for i, sample in enumerate(sampleset_field):
                for sample_field_name, sample_field in sample.items():
                    if sample_field_name == "vars":
                        for var_name, deci_var in sample_field.items():
                            for deci_var_field_name, deci_var_field in deci_var.items():
                                if deci_var_field_name == "value":
                                    index_list = []
                                    value_list = []
                                    for index, value in deci_var_field.items():
                                        index_list.append(
                                            list([np.array(i).tolist() for i in index])
                                        )
                                        value_list.append(np.array(value).tolist())
                                    obj["data"][i]["vars"][var_name]["value"] = [
                                        index_list,
                                        value_list,
                                    ]
    return obj


def from_old_sampleset(sampleset: typ.Union[SampleSet, jm.SampleSet]) -> SampleSet:
    if isinstance(sampleset, SampleSet):
        return sampleset

    solutions: list[dict[str, VariableSparseValue]] = []
    for label, _solutions in sampleset.record.solution.items():
        for i, sol in enumerate(_solutions):
            if len(solutions) <= i:
                solutions.append({})

            if isinstance(sol, np.ndarray):
                # If solution is dense type data format
                solutions[i][label] = VariableSparseValue.from_array(label, sol)

            else:
                # If solution is sparse type data format
                nnz_indices = sol[0]
                nnz_values = sol[1]
                shape = sol[2]

                if len(shape) == 0:
                    # sol is scalar
                    if len(nnz_values):
                        solutions[i][label] = VariableSparseValue.from_array(
                            label, np.array(nnz_values)[0]
                        )
                    else:
                        solutions[i][label] = VariableSparseValue.from_array(
                            label, np.array([])
                        )
                else:
                    # sol is array
                    nnz_inv_indices = np.array(list(nnz_indices)).T
                    solutions[i][label] = VariableSparseValue(
                        name=label,
                        value={
                            tuple(idx): value
                            for idx, value in zip(nnz_inv_indices, nnz_values)
                        },
                        shape=shape,
                        vartype=VarType.CONTINUOUS,
                    )

    data: list[Sample] = []
    # add same run_id for all samples
    run_id = str(uuid.uuid4())
    for i, sol in enumerate(solutions):
        energy: list[float] | None = sampleset.evaluation.energy
        objective: list[float] | None = sampleset.evaluation.objective
        const_violation = sampleset.evaluation.constraint_violations
        if energy is None:
            energy = []
        if objective is None:
            objective = []
        constraint: dict[str, ExprEvaluation] = {}
        if const_violation is not None:
            for label, const in const_violation.items():
                constraint[label] = ExprEvaluation(
                    name=label, total_violation=const[i], expr_values={}
                )
        data.append(
            Sample(
                vars=sol,
                evaluation_result=EvaluationResult(
                    energy=energy[i] if len(energy) else 0.0,
                    objective=objective[i] if len(objective) else 0.0,
                    constraints=constraint,
                    penalties={},
                ),
                # TODO: 1 -> sampleset.record.num_occurrences[i]
                num_occurrences=1,
                meta_info={},
                run_id=run_id,
            )
        )

    return SampleSet(
        data=data,
        set_info=sampleset.metadata,
        run_info={},
    )
